@@ -378,10 +378,10 @@ void ImmutableMessageFieldGenerator::GenerateKotlinDslMembers(
   printer->Print(variables_,
                  "$kt_deprecation$public var $kt_name$: $kt_type$\n"
                  "  @JvmName(\"${$get$kt_capitalized_name$$}$\")\n"
-                 "  get() = $kt_dsl_builder$.${$get$capitalized_name$$}$()\n"
+                 "  get() = $kt_dsl_builder$.${$$kt_safe_name$$}$\n"
                  "  @JvmName(\"${$set$kt_capitalized_name$$}$\")\n"
                  "  set(value) {\n"
-                 "    $kt_dsl_builder$.${$set$capitalized_name$$}$(value)\n"
+                 "    $kt_dsl_builder$.${$$kt_safe_name$$}$ = value\n"
                  "  }\n");
 
   WriteFieldAccessorDocComment(printer, descriptor_, CLEARER,
@@ -1376,7 +1376,7 @@ void RepeatedImmutableMessageFieldGenerator::GenerateKotlinDslMembers(
                  "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>\n"
                  "  @kotlin.jvm.JvmSynthetic\n"
                  "  get() = com.google.protobuf.kotlin.DslList(\n"
-                 "    $kt_dsl_builder$.${$get$capitalized_name$List$}$()\n"
+                 "    $kt_dsl_builder$.${$$kt_property_name$List$}$\n"
                  "  )\n");
 
   WriteFieldAccessorDocComment(printer, descriptor_, LIST_ADDER,
