@@ -405,7 +405,7 @@ bool HasWKTWithObjCCategory(const FileDescriptor* file) {
   // We don't check the name prefix or proto package because some files
   // (descriptor.proto), aren't shipped generated by the library, so this
   // seems to be the safest way to only catch the ones shipped.
-  const std::string name = file->name();
+  const absl::string_view name = file->name();
   if (name == "google/protobuf/any.proto" ||
       name == "google/protobuf/duration.proto" ||
       name == "google/protobuf/timestamp.proto") {
@@ -419,7 +419,7 @@ bool IsWKTWithObjCCategory(const Descriptor* descriptor) {
   if (!HasWKTWithObjCCategory(descriptor->file())) {
     return false;
   }
-  const std::string full_name = descriptor->full_name();
+  const absl::string_view full_name = descriptor->full_name();
   if (full_name == "google.protobuf.Any" ||
       full_name == "google.protobuf.Duration" ||
       full_name == "google.protobuf.Timestamp") {
