@@ -63,12 +63,30 @@ void RegisterFileLevelMetadata(const DescriptorTable* descriptor_table);
 
 }  // namespace internal
 
-using internal::DownCast;
 using internal::ReflectionOps;
 using internal::WireFormat;
 
 void Message::MergeImpl(MessageLite& to, const MessageLite& from) {
-  ReflectionOps::Merge(DownCast<const Message&>(from), DownCast<Message*>(&to));
+  ReflectionOps::Merge(DownCastMessage<Message>(from),
+                       DownCastMessage<Message>(&to));
+}
+
+void Message::ClearImpl(MessageLite& msg) {
+  ReflectionOps::Clear(&DownCastMessage<Message>(msg));
+}
+
+size_t Message::ByteSizeLongImpl(const MessageLite& msg) {
+  auto& _this = DownCastMessage<Message>(msg);
+  size_t size = WireFormat::ByteSize(_this);
+  _this.AccessCachedSize().Set(internal::ToCachedSize(size));
+  return size;
+}
+
+uint8_t* Message::_InternalSerializeImpl(const MessageLite& msg,
+                                         uint8_t* target,
+                                         io::EpsCopyOutputStream* stream) {
+  return WireFormat::_InternalSerialize(DownCastMessage<Message>(msg), target,
+                                        stream);
 }
 
 void Message::MergeFrom(const Message& from) {
@@ -81,10 +99,6 @@ void Message::MergeFrom(const Message& from) {
   }
 }
 
-void Message::CheckTypeAndMergeFrom(const MessageLite& other) {
-  MergeFrom(*DownCast<const Message*>(&other));
-}
-
 void Message::CopyFrom(const Message& from) {
   if (&from == this) return;
 
@@ -110,10 +124,12 @@ void Message::CopyFrom(const Message& from) {
   }
 }
 
+#if !defined(PROTOBUF_CUSTOM_VTABLE)
 void Message::Clear() { ReflectionOps::Clear(this); }
+#endif  // !PROTOBUF_CUSTOM_VTABLE
 
 bool Message::IsInitializedImpl(const MessageLite& msg) {
-  return ReflectionOps::IsInitialized(DownCast<const Message&>(msg));
+  return ReflectionOps::IsInitialized(DownCastMessage<Message>(msg));
 }
 
 void Message::FindInitializationErrors(std::vector<std::string>* errors) const {
@@ -155,6 +171,7 @@ Metadata Message::GetMetadataImpl(const ClassDataFull& data) {
   return {data.descriptor, data.reflection};
 }
 
+#if !defined(PROTOBUF_CUSTOM_VTABLE)
 uint8_t* Message::_InternalSerialize(uint8_t* target,
                                      io::EpsCopyOutputStream* stream) const {
   return WireFormat::_InternalSerialize(*this, target, stream);
@@ -165,6 +182,7 @@ size_t Message::ByteSizeLong() const {
   AccessCachedSize().Set(internal::ToCachedSize(size));
   return size;
 }
+#endif  // !PROTOBUF_CUSTOM_VTABLE
 
 size_t Message::ComputeUnknownFieldsSize(
     size_t total_size, internal::CachedSize* cached_size) const {
@@ -188,30 +206,33 @@ size_t Message::SpaceUsedLong() const {
   return GetClassData()->full().descriptor_methods->space_used_long(*this);
 }
 
-static std::string GetTypeNameImpl(const MessageLite& msg) {
-  return DownCast<const Message&>(msg).GetDescriptor()->full_name();
+absl::string_view Message::GetTypeNameImpl(const ClassData* data) {
+  return GetMetadataImpl(data->full()).descriptor->full_name();
 }
 
 static std::string InitializationErrorStringImpl(const MessageLite& msg) {
-  return DownCast<const Message&>(msg).InitializationErrorString();
+  return DownCastMessage<Message>(msg).InitializationErrorString();
 }
 
 const internal::TcParseTableBase* Message::GetTcParseTableImpl(
     const MessageLite& msg) {
-  return DownCast<const Message&>(msg).GetReflection()->GetTcParseTable();
+  return DownCastMessage<Message>(msg).GetReflection()->GetTcParseTable();
 }
 
 size_t Message::SpaceUsedLongImpl(const MessageLite& msg_lite) {
-  auto& msg = DownCast<const Message&>(msg_lite);
+  auto& msg = DownCastMessage<Message>(msg_lite);
   return msg.GetReflection()->SpaceUsedLong(msg);
 }
 
+static std::string DebugStringImpl(const MessageLite& msg) {
+  return DownCastMessage<Message>(msg).DebugString();
+}
+
 PROTOBUF_CONSTINIT const MessageLite::DescriptorMethods
     Message::kDescriptorMethods = {
-        GetTypeNameImpl,
-        InitializationErrorStringImpl,
-        GetTcParseTableImpl,
-        SpaceUsedLongImpl,
+        GetTypeNameImpl,     InitializationErrorStringImpl,
+        GetTcParseTableImpl, SpaceUsedLongImpl,
+        DebugStringImpl,
 };
 
 namespace internal {
@@ -456,9 +477,11 @@ const internal::RepeatedFieldAccessor* Reflection::RepeatedFieldAccessor(
     HANDLE_PRIMITIVE_TYPE(ENUM, int32_t)
 #undef HANDLE_PRIMITIVE_TYPE
     case FieldDescriptor::CPPTYPE_STRING:
-      switch (field->options().ctype()) {
-        default:
-        case FieldOptions::STRING:
+      switch (field->cpp_string_type()) {
+        case FieldDescriptor::CppStringType::kCord:
+          ABSL_LOG(FATAL) << "Repeated cords are not supported.";
+        case FieldDescriptor::CppStringType::kView:
+        case FieldDescriptor::CppStringType::kString:
           return GetSingleton<internal::RepeatedPtrFieldStringAccessor>();
       }
       break;
